#include "alloc.h"
#include "exception.h"
#include "gdt.h"
#include "idt.h"
#include "interrupt.h"
#include "io.h"
#include "irq.h"
#include "klibc.h"
#include "kthread.h"
#include "mem.h"
#include "multiboot.h"
#include "paging.h"
#include "pit.h"
#include "serial.h"
#include "stack.h"
#include "stdarg.h"
#ifdef RUN_TEST
#include "test.h"
#endif
#include "time.h"
#include "types.h"
#include "vga.h"

#define CHECK_FLAG(flags, bit) ((flags) & (1 << (bit)))

void idleThread(void *arg)
{
    (void)arg;
    while (1) {
        printIntDetails((jiffies / HZ), GREEN, BLACK, 0, VGA_HEIGHT - 1);
        kthreadYield();
    }
}

// Multiboot information available here :
// https://www.gnu.org/software/grub/manual/multiboot/multiboot.html#kernel_002ec
// https://www.gnu.org/software/grub/manual/multiboot/html_node/Boot-information-format.html#Boot%20information%20format
void kmain(unsigned long magic, unsigned long addr)
{
    unsigned long upperMemKB = 0;
    int memMapAvailable = 0;
    paddr_t lastUsedByMem;

    VGASetup(BLACK, GREEN);
    cursorEnable(14, 15);

    printf("Setting up Interruptions\n");
    gdtSetup();
    idtSetup();
    irqSetup();
    pitSetup(HZ);

    if (magic == MULTIBOOT_BOOTLOADER_MAGIC) { // Get loaded by Grub with mutliboot version 1
        multiboot_info_t *mbi = (multiboot_info_t *)addr;
        /* Are mem_* valid? */
        if (CHECK_FLAG(mbi->flags, 0)) {
            printf("mem_lower = %dKiB mem_upper %dKiB\n", mbi->mem_lower, mbi->mem_upper);
            upperMemKB = mbi->mem_upper;
        }

        /* Is boot_device valid? */
        if (CHECK_FLAG(mbi->flags, 1)) {
            printf("boot_device = %d\n", mbi->boot_device);
        }

        /* Is the command line passed? */
        if (CHECK_FLAG(mbi->flags, 2)) {
            printf("cmdline = %s\n", (char *)mbi->cmdline);
        }

        if (CHECK_FLAG(mbi->flags, 3)) {
            multiboot_module_t *mod;
            uint32_t i;

            printf("mods_count = %d, mods_addr = 0x%x\n", (int)mbi->mods_count,
                   (int)mbi->mods_addr);
            for (i = 0, mod = (multiboot_module_t *)mbi->mods_addr; i < mbi->mods_count;
                 i++, mod++)
                printf(" mod_start = 0x%x, mod_end = 0x%x, cmdline = %s\n",
                       (unsigned)mod->mod_start, (unsigned)mod->mod_end, (char *)mod->cmdline);
        }

        if (CHECK_FLAG(mbi->flags, 6)) {
            memMapAvailable = 1;
        }
    }

    if (upperMemKB == 0) {
        printf("Cannot get upper phy mem bound. Using default value 32MB\n");
        upperMemKB = 32 * 1024;
    }

    printf("Setting up Mem\n");
    memSetup(upperMemKB, &lastUsedByMem);

    if (memMapAvailable) {
        multiboot_info_t *mbi             = (multiboot_info_t *)addr;
        struct multiboot_mmap_entry *mmap = (struct multiboot_mmap_entry *)mbi->mmap_addr;
        uint size = mbi->mmap_length / sizeof(struct multiboot_mmap_entry);
        pr_devel("mmap buffer at 0x%x with %d entries\n", mbi->mmap_addr, size);
        for (uint i = 0; i < size; i++) {
            printf("   base_addr 0x%llx, length = 0x%llx, type = 0x%x\n", mmap[i].addr,
                   mmap[i].len, (uint32_t)mmap[i].type);
            memAddBank(
                max(mmap[i].addr, (multiboot_uint64_t)lastUsedByMem),
                min((multiboot_uint64_t)(upperMemKB * 1024), mmap[i].addr + mmap[i].len),
                mmap[i].type == MULTIBOOT_MEMORY_AVAILABLE);
        }
    } else {
        printf("Cannot get memory Mapping information, using default value\n");
        memAddBank(0, lastUsedByMem, 0);
        memAddBank(lastUsedByMem, upperMemKB * 1024, 1);
    }
#ifdef RUN_TEST
    testPhymem();
#endif
    printf("Setting up Pagination\n");
    pagingSetup(lastUsedByMem);

    printf("Setting up IRQ handlers\n");
    irqSetRoutine(IRQ_KEYBOARD, keyboard_handler);

    printf("Enabling HW interrupts\n");
    // Enabling the HW interrupts
    exceptionSetup();
    asm volatile("sti\n");

    printf("Setting up Serial link (115200)\n");
    serialSetup(115200);
    allocSetup();
    kthreadSetup(_stack_bottom, (_stack_top - _stack_bottom + 1));
    kthreadCreate("idle ", idleThread, NULL);
    irqSetRoutine(IRQ_TIMER, pit_handler);
#ifdef RUN_TEST
    run_test();
#endif
    printf("\nSystem init done\n");

    // There is no real caller behind this point
    // So finish this by ourself
    kthreadExit();
}
